'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var puppeteer = _interopDefault(require('puppeteer'));

function stringifySelectors ({matches, children, html}, options) {
  const result = {
    matches: matches.map(match => {
      const result = {
        ...match,
        selector: match.selector.map(part => {
          const [unmatched, matched] = options.formatSelector(...part);
          return `${unmatched} ${matched}`.trim().replace(/\s+/g, ' ')
        }).join(', ')
      };
      return result
    })
  };
  if (html) {
    result.html = html;
  }
  if (children) {
    result.children = children.map(child => stringifySelectors(child, options));
  }
  return result
}

async function createPage (browser, styles, html) {
  const page = await browser.newPage();
  await page.setContent(html);
  for (let style of styles) {
    await page.addStyleTag(style);
  }
  page.on('console', msg => console.log(msg.text()));
  return page
}
function getElementQuery (html) {
  const htmlWithNoComments = html.replace(/<!--[\s\S]*?-->/g, '');
  const match = /^\s*<\s*([a-z]+)/i.exec(htmlWithNoComments);
  if (match) {
    return `${match[1].toLowerCase()}:first-of-type`
  }
  throw new Error('Input HTML does not contain a valid tag.')
}
function findMatchingRules (elementQuery, options) {
  function getCssRules(styles) {
    const CSS_RULE_TYPES = [
      'UNKNOWN_RULE',
      'STYLE_RULE',
      'CHARSET_RULE',
      'IMPORT_RULE',
      'MEDIA_RULE',
      'FONT_FACE_RULE',
      'PAGE_RULE',
      'KEYFRAMES_RULE',
      'KEYFRAME_RULE',
      null,
      'NAMESPACE_RULE',
      'COUNTER_STYLE_RULE',
      'SUPPORTS_RULE',
      'DOCUMENT_RULE',
      'FONT_FEATURE_VALUES_RULE',
      'VIEWPORT_RULE',
      'REGION_STYLE_RULE'
    ];
    const rules = [];
    for (const {cssRules} of styles) {
      for (const rule of cssRules) {
        switch (CSS_RULE_TYPES[rule.type]) {
          case 'STYLE_RULE':
            rules.push(rule);
            break
          case 'MEDIA_RULE':
            rules.push(...getCssRules([rule]));
            break
        }
      }
    }
    return rules
  }
  function stringifyElement(element) {
    const match = element.outerHTML.match(/[^>]*>/);
    return match ? match[0] : ''
  }
  function findRulesForElement(matches, rules, element, options, depth) {
    const result = {
      matches: rules.reduce((acc, rule) => {
        let hasMatch = false;
        const selector = rule.selectorText.split(/\s*,\s*/).map(part => {
          let parsed;
          if (options.includePartialMatches) {
            parsed = findMatchingPartOfSelector(matches, element, part, depth);
          } else {
            parsed = testIfSelectorIsMatch(matches, element, part);
          }
          if (parsed[1]) {
            hasMatch = true;
          }
          return parsed
        });
        if (hasMatch) {
          acc.push(formatRule(selector, rule, options));
        }
        return acc
      }, [])
    };
    if (options.includeHtml === true) {
      result.html = stringifyElement(element);
    }
    if (options.recursive === true) {
      const depthOfChildren = depth + 1;
      result.children = Array.prototype.map.call(element.children, child => {
        return findRulesForElement(matches, rules, child, options, depthOfChildren)
      });
    }
    return result
  }
  function testIfSelectorIsMatch(matches, element, selector) {
    if (matches(element, selector)) {
      return ['', selector]
    }
    return [selector, '']
  }
  function findMatchingPartOfSelector(matches, element, selector, depth) {
    const parts = selector.split(/\s+/);
    for (let i = 0, part = parts[i]; part; part = parts[++i]) {
      if (/[>+~]/.test(part)) {
        if (combinatorPreventsMatch(matches, element, parts, i, depth)) {
          break
        }
        continue
      }
      const matched = parts.slice(i).join(' ');
      if (matches(element, matched)) {
        const unmatched = parts.slice(0, i).join(' ');
        return [unmatched, matched]
      }
    }
    return [selector, '']
  }
  function combinatorPreventsMatch(matches, element, parts, index, elementDepth) {
    if (elementDepth < 1) {
      return false
    }
    if (selectorHasDescendentCombinator(parts, index)) {
      return false
    }
    let depthDiff = elementDepth;
    for (let i = index + 2; i < parts.length; i++) {
      if (parts[i] === '>') {
        depthDiff--;
      }
    }
    if (depthDiff < 1) {
      return false
    }
    const selector = parts.slice(0, index).join(' ');
    const {elements, depth} = getElementsUsingCombinator(element, parts[index], elementDepth);
    return !elements.some(node => findMatchingPartOfSelector(matches, node, selector, depth)[1])
  }
  function selectorHasDescendentCombinator(parts, index) {
    for (let i = index + 1; i < parts.length - 1; i++) {
      if (/[>+~]/.test(parts[i + 1])) {
        i++;
      } else {
        return true
      }
    }
    return false
  }
  function getElementsUsingCombinator(element, combinator, depth) {
    const elements = [];
    let depthOfElements = depth;
    if (combinator === '>') {
      if (element.parentElement) {
        elements.push(element.parentElement);
      }
      depthOfElements--;
    } else if (combinator === '+') {
      if (element.previousElementSibling) {
        elements.push(element.previousElementSibling);
      }
    } else if (combinator === '~') {
      let el = element;
      while ((el = el.previousElementSibling)) {
        elements.unshift(el);
      }
    }
    return {elements, depth: depthOfElements}
  }
  function cssTextToArray(cssText) {
    const match = cssText.match(/{([^}]*)}/);
    const text = match ? match[1].trim() : '';
    return text.split(/;\s*/).reduce((acc, str) => {
      str && acc.push(`${str}`);
      return acc
    }, [])
  }
  function formatRule(selector, rule, options) {
    const ruleObj = {selector};
    const media = getMediaText(rule);
    if (media) {
      ruleObj.media = media;
    }
    if (options.includeCss === true) {
      ruleObj.css = cssTextToArray(rule.cssText);
    }
    if (options.includePartialMatches) {
      ruleObj.isPartialMatch = selector.every(([unmatched]) => unmatched);
    }
    return ruleObj
  }
  function getMediaText(rule) {
    const media = [];
    let current = rule;
    while ((current = current.parentRule) && current.media) {
      media.push(current.media.mediaText);
    }
    return media.join(' AND ')
  }
  const matches = Function.call.bind(window.Element.prototype.webkitMatchesSelector);
  const rules = getCssRules(document.styleSheets);
  let element = document.querySelector(elementQuery);
  element = element.parentNode.removeChild(element);
  return findRulesForElement(matches, rules, element, options, 0)
}
async function findMatchesFromPage (styles, html, options) {
  const elementQuery = getElementQuery(html);
  const browser = await puppeteer.launch();
  let selectors;
  try {
    const page = await createPage(browser, styles, html);
    selectors = await page.evaluate(findMatchingRules, elementQuery, options);
    selectors = stringifySelectors(selectors, options);
  } catch (error) {
    browser.close();
    throw error
  }
  browser.close();
  return selectors
}

const DEFAULT_OPTIONS = {
  recursive: true,
  includeHtml: false,
  includeCss: false,
  includePartialMatches: true,
  formatSelector: (a, b) => [a, b]
};
function normalizeStyles (styles) {
  if (Array.isArray(styles)) {
    return styles
  } else if (typeof styles === 'string') {
    return [{content: styles}]
  }
  return [styles]
}
function findMatches (styles, html, userOptions) {
  const stylesArray = normalizeStyles(styles);
  const options = Object.assign({}, DEFAULT_OPTIONS, userOptions);
  return findMatchesFromPage(stylesArray, html, options)
}

exports.findMatches = findMatches;
